from typing import Dict, Set

from pegen.first_sets import FirstSetCalculator
from pegen.grammar import Grammar
from pegen.grammar_parser import GeneratedParser as GrammarParser
from pegen.utils import parse_string


def calculate_first_sets(grammar_source: str) -> Dict[str, Set[str]]:
    grammar: Grammar = parse_string(grammar_source, GrammarParser)
    return FirstSetCalculator(grammar.rules).calculate()


def test_alternatives() -> None:
    grammar = """
        start: expr NEWLINE? ENDMARKER
        expr: A | B
        A: 'a' | '-'
        B: 'b' | '+'
    """
    assert calculate_first_sets(grammar) == {
        "A": {"'a'", "'-'"},
        "B": {"'+'", "'b'"},
        "expr": {"'+'", "'a'", "'b'", "'-'"},
        "start": {"'+'", "'a'", "'b'", "'-'"},
    }


def test_optionals() -> None:
    grammar = """
        start: expr NEWLINE
        expr: ['a'] ['b'] 'c'
    """
    assert calculate_first_sets(grammar) == {
        "expr": {"'c'", "'a'", "'b'"},
        "start": {"'c'", "'a'", "'b'"},
    }


def test_repeat_with_separator() -> None:
    grammar = """
    start: ','.thing+ NEWLINE
    thing: NUMBER
    """
    assert calculate_first_sets(grammar) == {"thing": {"NUMBER"}, "start": {"NUMBER"}}


def test_optional_operator() -> None:
    grammar = """
    start: sum NEWLINE
    sum: (term)? 'b'
    term: NUMBER
    """
    assert calculate_first_sets(grammar) == {
        "term": {"NUMBER"},
        "sum": {"NUMBER", "'b'"},
        "start": {"'b'", "NUMBER"},
    }


def test_optional_literal() -> None:
    grammar = """
    start: sum NEWLINE
    sum: '+' ? term
    term: NUMBER
    """
    assert calculate_first_sets(grammar) == {
        "term": {"NUMBER"},
        "sum": {"'+'", "NUMBER"},
        "start": {"'+'", "NUMBER"},
    }


def test_optional_after() -> None:
    grammar = """
    start: term NEWLINE
    term: NUMBER ['+']
    """
    assert calculate_first_sets(grammar) == {"term": {"NUMBER"}, "start": {"NUMBER"}}


def test_optional_before() -> None:
    grammar = """
    start: term NEWLINE
    term: ['+'] NUMBER
    """
    assert calculate_first_sets(grammar) == {"term": {"NUMBER", "'+'"}, "start": {"NUMBER", "'+'"}}


def test_repeat_0() -> None:
    grammar = """
    start: thing* "+" NEWLINE
    thing: NUMBER
    """
    assert calculate_first_sets(grammar) == {"thing": {"NUMBER"}, "start": {'"+"', "NUMBER"}}


def test_repeat_0_with_group() -> None:
    grammar = """
    start: ('+' '-')* term NEWLINE
    term: NUMBER
    """
    assert calculate_first_sets(grammar) == {"term": {"NUMBER"}, "start": {"'+'", "NUMBER"}}


def test_repeat_1() -> None:
    grammar = """
    start: thing+ '-' NEWLINE
    thing: NUMBER
    """
    assert calculate_first_sets(grammar) == {"thing": {"NUMBER"}, "start": {"NUMBER"}}


def test_repeat_1_with_group() -> None:
    grammar = """
    start: ('+' term)+ term NEWLINE
    term: NUMBER
    """
    assert calculate_first_sets(grammar) == {"term": {"NUMBER"}, "start": {"'+'"}}


def test_gather() -> None:
    grammar = """
    start: ','.thing+ NEWLINE
    thing: NUMBER
    """
    assert calculate_first_sets(grammar) == {"thing": {"NUMBER"}, "start": {"NUMBER"}}


def test_positive_lookahead() -> None:
    grammar = """
    start: expr NEWLINE
    expr: &'a' opt
    opt: 'a' | 'b' | 'c'
    """
    assert calculate_first_sets(grammar) == {
        "expr": {"'a'"},
        "start": {"'a'"},
        "opt": {"'b'", "'c'", "'a'"},
    }


def test_negative_lookahead() -> None:
    grammar = """
    start: expr NEWLINE
    expr: !'a' opt
    opt: 'a' | 'b' | 'c'
    """
    assert calculate_first_sets(grammar) == {
        "opt": {"'b'", "'a'", "'c'"},
        "expr": {"'b'", "'c'"},
        "start": {"'b'", "'c'"},
    }


def test_left_recursion() -> None:
    grammar = """
    start: expr NEWLINE
    expr: ('-' term | expr '+' term | term)
    term: NUMBER
    foo: 'foo'
    bar: 'bar'
    baz: 'baz'
    """
    assert calculate_first_sets(grammar) == {
        "expr": {"NUMBER", "'-'"},
        "term": {"NUMBER"},
        "start": {"NUMBER", "'-'"},
        "foo": {"'foo'"},
        "bar": {"'bar'"},
        "baz": {"'baz'"},
    }


def test_advance_left_recursion() -> None:
    grammar = """
    start: NUMBER | sign start
    sign: ['-']
    """
    assert calculate_first_sets(grammar) == {"sign": {"'-'", ""}, "start": {"'-'", "NUMBER"}}


def test_mutual_left_recursion() -> None:
    grammar = """
    start: foo 'E'
    foo: bar 'A' | 'B'
    bar: foo 'C' | 'D'
    """
    assert calculate_first_sets(grammar) == {
        "foo": {"'D'", "'B'"},
        "bar": {"'D'"},
        "start": {"'D'", "'B'"},
    }


def test_nasty_left_recursion() -> None:
    # TODO: Validate this
    grammar = """
    start: target '='
    target: maybe '+' | NAME
    maybe: maybe '-' | target
    """
    assert calculate_first_sets(grammar) == {"maybe": set(), "target": {"NAME"}, "start": {"NAME"}}


def test_nullable_rule() -> None:
    grammar = """
    start: sign thing $
    sign: ['-']
    thing: NUMBER
    """
    assert calculate_first_sets(grammar) == {
        "sign": {"", "'-'"},
        "thing": {"NUMBER"},
        "start": {"NUMBER", "'-'"},
    }


def test_epsilon_production_in_start_rule() -> None:
    grammar = """
    start: ['-'] $
    """
    assert calculate_first_sets(grammar) == {"start": {"ENDMARKER", "'-'"}}


def test_multiple_nullable_rules() -> None:
    grammar = """
    start: sign thing other another $
    sign: ['-']
    thing: ['+']
    other: '*'
    another: '/'
    """
    assert calculate_first_sets(grammar) == {
        "sign": {"", "'-'"},
        "thing": {"'+'", ""},
        "start": {"'+'", "'-'", "'*'"},
        "other": {"'*'"},
        "another": {"'/'"},
    }



"""Test identifying unsupported syntax construction in older Python versions.

Note that we can request the parser to apply stricter bounds on the parsing but
not broader since we would not be able to generate the proper ast nodes.

"""
import io
import tokenize

import pytest
from pegen.tokenizer import Tokenizer


# matrix mul 3.5
@pytest.mark.parametrize("source", ["a @ b", "a @= b"])
def test_mat_mult(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 4))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "The '@' operator is" in e.exconly()


# await 3.5
def test_await(python_parser_cls):
    temp = io.StringIO("await b")
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 4))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Await expressions are" in e.exconly()


# async 3.5
@pytest.mark.parametrize(
    "source, message",
    [
        ("async def f():\n    pass", "Async functions are"),
        ("async with a:\n    pass", "Async with statements are"),
        ("async for a in b:\n    pass", "Async for loops are"),
    ],
)
def test_async(python_parser_cls, source, message):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 4))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert message in e.exconly()


# async comprehension 3.6
def test_async_comprehension(python_parser_cls):
    temp = io.StringIO("""[a async for a in b if c]""")
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 5))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")
    assert "Async comprehensions are" in e.exconly()


# variable annotation 3.6
@pytest.mark.parametrize("source", ["a: int = 1", "(a): int "])
def test_variable_annotation(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 5))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Variable annotation syntax is" in e.exconly()


# pos only args 3.8
@pytest.mark.parametrize("source", ["def f(a,/):\n\tpass", "def f(a=1,/):\n\tpass"])
def test_pos_only_args(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 7))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Positional only arguments are" in e.exconly()


# assignment operator 3.8
@pytest.mark.parametrize("source", ["a := 1"])
def test_assignment_operator(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 7))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "The ':=' operator is" in e.exconly()


# generic decorators 3.9
@pytest.mark.parametrize("source", ["@f[1]\ndef f():\n\tpass"])
def test_generic_decorators(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 8))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Generic decorator are" in e.exconly()


# parenthesized with items 3.9
@pytest.mark.parametrize("source", ["with (a, b):\n\tpass"])
def test_parenthesized_with_items(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 8))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Parenthesized with items" in e.exconly()


# match 3.10
@pytest.mark.parametrize(
    "source", ["match a:\n\tcase 1:\n\t\tpass", "match a", "match a:\ncase b"]
)
def test_match_statement(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 9))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Pattern matching is" in e.exconly()


# try except * 3.11
@pytest.mark.parametrize("source", ["try:\n\ta = 1\nexcept *ValueError:\n\tpass"])
def test_exceptgroup_statement(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 10))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Exception groups are" in e.exconly()


# type alias and type vars 3.12
@pytest.mark.parametrize("source", ["type T = int", "type T[U] = tuple[U]"])
def test_type_params_statement(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 11))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Type statement is" in e.exconly() or "Type parameter lists are" in e.exconly()


# type alias and type vars 3.12
@pytest.mark.parametrize("source", ["def f[T]():\n\tpass", "async def f[T]():\n\tpass"])
def test_generic_function_statement(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 11))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Type parameter lists are" in e.exconly()


# generic classes 3.12
@pytest.mark.parametrize("source", ["class A[T]:\n\tpass"])
def test_generic_class_statement(python_parser_cls, source):
    temp = io.StringIO(source)
    tokengen = tokenize.generate_tokens(temp.readline)
    tokenizer = Tokenizer(tokengen, verbose=False)
    pp = python_parser_cls(tokenizer, py_version=(3, 11))
    with pytest.raises(SyntaxError) as e:
        pp.parse("file")

    assert "Type parameter lists are" in e.exconly()



a = (k for k in g)
b = (k for k in g if k == 1)
(k for k in g).send(None)


a = [k for k in g]
b = [k for k in g if k == 1]


a = {k for k in g}
b = {k for k in g if k == 1}
a = {k: 1 for k in g}
b = {k: 2 for k in g if k == 1}


[k for v in a for k in v]



try:
    try:
        raise ExceptionGroup(
            "eg", [TypeError(1), ValueError(2), OSError(3)])
    except* TypeError as e:
        raise
    except* ValueError as e:
        raise
    # OSError not handled
except ExceptionGroup as e:
    exc = e



try:
    try:
        raise ExceptionGroup(
            "eg", [TypeError(1), ValueError(2), OSError(3)])
    except* TypeError:
        raise
    except* ValueError:
        raise
    # OSError not handled
except ExceptionGroup as e:
    exc = e



try:
    try:
        raise ExceptionGroup(
            "eg", [TypeError(1), ValueError(2), OSError(3)])
    except* TypeError as e:
        raise
    except* ValueError as e:
        pass
    # OSError not handled
except ExceptionGroup as e:
    exc = e



try:
    try:
        raise ExceptionGroup(
            "eg", [TypeError(1), ValueError(2)])
    except* TypeError:
        raise
    except* ValueError:
        pass
except ExceptionGroup as e:
    exc = e



try:
    try:
        raise ExceptionGroup(
            "eg", [TypeError(1), ValueError(2), OSError(3)])
    except* TypeError as e:
        raise
    except* ValueError as e:
        pass
    # OSError not handled
except ExceptionGroup as e:
    exc = e



try:
    try:
        raise ExceptionGroup(
            "eg", [TypeError(1), ValueError(2), OSError(3)])
    except* TypeError:
        raise
    except* ValueError:
        pass
except ExceptionGroup as e:
    exc = e



try:
    try:
        raise ValueError(42)
    except* ValueError as e:
        raise
except ExceptionGroup as e:
    exc = e



try:
    try:
        raise ValueError(42)
    except* ValueError:
        raise
except ExceptionGroup as e:
    exc = e



orig = ExceptionGroup("eg", [ValueError(1), OSError(2)])
try:
    try:
        raise orig
    except* OSError as e:
        raise TypeError(3)
except ExceptionGroup as e:
    exc = e



orig = ExceptionGroup("eg", [ValueError(1), OSError(2)])
try:
    try:
        raise orig
    except* OSError:
        raise TypeError(3)
except ExceptionGroup as e:
    exc = e



orig = ExceptionGroup("eg", [TypeError(1), ValueError(2)])
try:
    try:
        raise orig
    except* (TypeError, ValueError) as e:
        raise SyntaxError(3)
except SyntaxError as e:
    exc = e



orig = ExceptionGroup("eg", [TypeError(1), ValueError(2)])
try:
    try:
        raise orig
    except* (TypeError, ValueError) as e:
        raise SyntaxError(3)
except SyntaxError as e:
    exc = e



orig = ExceptionGroup("eg", [TypeError(1), ValueError(2)])
try:
    try:
        raise orig
    except* TypeError as e:
        raise SyntaxError(3)
    except* ValueError as e:
        raise SyntaxError(4)
except ExceptionGroup as e:
    exc = e



orig = ExceptionGroup("eg", [TypeError(1), ValueError(2)])
try:
    try:
        raise orig
    except* TypeError:
        raise SyntaxError(3)
    except* ValueError:
        raise SyntaxError(4)
except ExceptionGroup as e:
    exc = e



orig = ExceptionGroup("eg", [ValueError(1), OSError(2)])
try:
    try:
        raise orig
    except* OSError as e:
        raise TypeError(3) from e
except ExceptionGroup as e:
    exc = e




orig = ExceptionGroup("eg", [ValueError(1), OSError(2)])
try:
    try:
        raise orig
    except* OSError:
        e = sys.exception()
        raise TypeError(3) from e
except ExceptionGroup as e:
    exc = e



orig = ExceptionGroup("eg", [TypeError(1), ValueError(2)])
try:
    try:
        raise orig
    except* (TypeError, ValueError) as e:
        raise SyntaxError(3) from e
except SyntaxError as e:
    exc = e



orig = ExceptionGroup("eg", [TypeError(1), ValueError(2)])
try:
    try:
        raise orig
    except* (TypeError, ValueError) as e:
        e = sys.exception()
        raise SyntaxError(3) from e
except SyntaxError as e:
    exc = e



orig = ExceptionGroup("eg", [TypeError(1), ValueError(2)])
try:
    try:
        raise orig
    except* TypeError as e:
        raise SyntaxError(3) from e
    except* ValueError as e:
        raise SyntaxError(4) from e
except ExceptionGroup as e:
    exc = e



orig = ExceptionGroup("eg", [TypeError(1), ValueError(2)])
try:
    try:
        raise orig
    except* TypeError:
        e = sys.exception()
        raise SyntaxError(3) from e
    except* ValueError:
        e = sys.exception()
        raise SyntaxError(4) from e
except ExceptionGroup as e:
    exc = e


